package com.sillycycle.bagleyd.tertris;

/*
 * @(#)PolyiamondImage.java
 *
 * Copyright 1993 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;

/**
 * The <code>PolyiamondImage</code> class holds the colors and
 * handles the construction of the images for the different
 * polyiamonds.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class PolyiamondImage extends Component {
	private static final long serialVersionUID = 42L;
	boolean plainPiece = false;

	static String[] colorStrings = { // 113
		"Red", "Green", "Blue", "Yellow",
		"Magenta", "Cyan", "FireBrick", "GreenYellow",
		"SlateBlue", "Khaki", "Plum", "Violet",
		"DarkTurquoise", "Gold", "Orchid", "Turquoise",
		"Orange", "OrangeRed", "VioletRed", "BlueViolet",
		"SeaGreen", "Pink", "ForestGreen", "SkyBlue",
		"Coral", "Wheat", "Goldenrod", "IndianRed",
		"SpringGreen", "CornflowerBlue", "Thistle", "Aquamarine",
		"CadetBlue", "LightSteelBlue", "NavyBlue", "SteelBlue",
		"YellowGreen", "DarkViolet", "MediumSeaGreen", "DarkSlateGray",
		"White", "MediumVioletRed", "Sienna", "MediumAquamarine",
		"MediumBlue", "Navy", "DarkOliveGreen", "DarkGreen",
		"DarkOliveGreen4", "Tan", "MediumTurquoise", "DarkSlateBlue",
		"Maroon", "MediumSlateBlue", "PaleGoldenrod", "MediumSpringGreen",
		"DarkOrchid", "LightBlue", "MidnightBlue", "LimeGreen",
		"PaleGreen", "DarkGoldenrod", "DarkSeaGreen", "MediumOrchid",
		"salmon", "brown", "SandyBrown", "SaddleBrown",
		"RoyalBlue", "RosyBrown", "PowderBlue", "PeachPuff",
		"PaleTurquoise", "PaleGoldenrod", "OliveDrab", "LawnGreen",
		"HotPink", "DodgerBlue", "DeepSkyBlue", "DeepPink",
		"DarkSalmon", "DarkOrange", "DarkKhaki", "AliceBlue",
		"MediumPurple", "bisque", "burlywood", "chartreuse",
		"chocolate", "moccasin", "peru", "purple",
		"tomato", "LightYellow4", "LightSalmon4", "LavenderBlush4",
		"LightSkyBlue4", "MistyRose4", "NavajoWhite4", "MediumOrchid4",
		"RosyBrown4", "VioletRed4", "PeachPuff4", "RoyalBlue4",
		"SeaGreen4", "OliveDrab4", "PaleTurquoise4", "PaleVioletRed4",
		"LightPink4", "DeepPink4", "DarkGoldenrod4", "CadetBlue4",
		"DarkSeaGreen4"
	};

	static Color[] colorThings = {
		Color.red, Color.green, Color.blue, Color.yellow,
		Color.magenta, Color.cyan,
			new Color(178, 34, 34), new Color(173, 255, 47),
		new Color(131, 111, 255), new Color(255, 246, 143),
			new Color(255, 187, 255), new Color(238, 130, 238),
		new Color(0, 206, 209), new Color(255, 215, 0),
			new Color(218, 112, 214), new Color(0, 229, 238),
		new Color(255, 165, 0), new Color(255, 69, 0),
			new Color(208, 32, 144), new Color(138, 43, 226),
		new Color(78, 238, 148), new Color(238, 169, 184),
			new Color(34, 139, 34), new Color(126, 192, 238),
		new Color(255, 127, 80), new Color(245, 222, 179),
			new Color(218, 165, 32), new Color(205, 92, 92),
		new Color(0, 238, 118), new Color(100, 149, 237),
			new Color(216, 191, 216), new Color(127, 255, 212),
		new Color(95, 158, 160), new Color(176, 196, 222),
			new Color(0, 0, 128), new Color(70, 130, 180),
		new Color(154, 205, 50), new Color(148, 0, 211),
			new Color(60, 179, 113), new Color(151, 255, 255),
		new Color(110, 139, 61), new Color(199, 21, 133),
			new Color(238, 121, 66), new Color(102, 205, 170),
		new Color(0, 0, 208), new Color(0, 0, 177),
			new Color(85, 107, 47), new Color(0, 100, 0),
		new Color(170, 170, 170), new Color(210, 180, 140),
			new Color(72, 209, 204), new Color(72, 61, 139),
		new Color(176, 48, 96), new Color(123, 104, 238),
			new Color(238, 232, 170), new Color(0, 250, 154),
		new Color(153, 50, 204), new Color(173, 216, 230),
			new Color(25, 25, 112), new Color(50, 205, 50),
		new Color(152, 251, 152), new Color(184, 134, 11),
			new Color(143, 188, 143), new Color(186, 85, 211),
		new Color(250, 128, 114), new Color(165, 42, 42),
			new Color(244, 164, 96), new Color(139, 69, 19),
		new Color(65, 105, 225), new Color(188, 143, 143),
			new Color(176, 224, 230), new Color(255, 218, 185),
		new Color(175, 238, 238), new Color(238, 232, 170),
			new Color(107, 142, 35), new Color(124, 252, 0),
		new Color(255, 105, 180), new Color(30, 144, 255),
			new Color(0, 191, 255), new Color(255, 20, 147),
		new Color(233, 150, 122), new Color(255, 140, 0),
			new Color(189, 183, 107), new Color(240, 248, 255),
		new Color(147, 112, 219), new Color(255, 228, 196),
			new Color(222, 184, 135), new Color(127, 255, 0),
		new Color(210, 105, 30), new Color(255, 228, 181),
			new Color(205, 133, 63), new Color(160, 32, 240),
		new Color(255, 99, 71), new Color(139, 139, 122),
			new Color(139, 87, 66), new Color(139, 131, 134),
		new Color(96, 123, 139), new Color(139, 125, 123),
			new Color(139, 121, 94), new Color(122, 55, 139),
		new Color(139, 105, 105), new Color(139, 34, 82),
			new Color(139, 119, 101), new Color(39, 64, 139),
		new Color(105, 139, 105), new Color(105, 139, 34),
			new Color(102, 139, 139), new Color(139, 71, 93),
		new Color(139, 95, 101), new Color(139, 10, 80),
			new Color(139, 101, 8), new Color(83, 134, 139),
		new Color(105, 139, 105)
	};

	static final int NUM_BITMAPS = 8;
	static final int DOWN = 1; /* (1<<0) */ /* Down side of Up Triangle */
	static final int UP = 4; /* (1<<2) */ /* Up side of Down Triangle */
	static final int MAX_ORIENT = 2;
	static final int DELTAX = 6;
	static final int DELTAY = 3;
	static final double SQRT3 = 1.732050807569;
	Color background;
	int pm;

	/* These need to be written better when there is a bitmap */
	static int checkDown(int x) {
		return (x | DOWN);
	}
	static int checkUp(int x) {
		return (x | UP);
	}
	static int swapUpDown(int x) {
		return (((x&4)>>2)+(x&2)+((x&1)<<2));
	}

	Point[][] triangleUnit = new Point[MAX_ORIENT][3];

	public static int hexCol(int i) {
		return 4 * i - 1;
	}

	public static int hexRow(int j) {
		return 2 * j;
	}

	public static boolean inHexagon(int size, int i, int j) {
		int w = hexCol(size), h = hexRow(size);
		int newI = i, newJ = j;

		if (newI > 2 * size - 1)
			newI = w - 1 - newI;
		if (newJ >= size)
			newJ = h - 1 - newJ;
		return (newI + newJ >= size - 1);
	}

	public PolyiamondImage(Color color) {
		this.background = color;
		triangleUnit[0][0] = new Point(0, 0);
		triangleUnit[0][1] = new Point(-1, -1);
		triangleUnit[0][2] = new Point(2, 0);
		/* triangleUnit[0][3] = new Point(-1, 1); */
		triangleUnit[1][0] = new Point(0, 0);
		triangleUnit[1][1] = new Point(1, 1);
		triangleUnit[1][2] = new Point(-2, 0);
		/* triangleUnit[1][3] = new Point(1, -1); */
	}

	@Override
	public void setBackground(Color color) {
		this.background = color;
	}

	static int polyPixmap(int i, int j, int[][] shape) {
		return NUM_BITMAPS - 1 - (shape[j][i] % NUM_BITMAPS);
	}

	void drawUnit(Graphics g, int pmid, Color cid, Color bid,
			boolean invert,
			int x, int y, int i, int j, int width, int height,
			boolean piece) {
		int[] triangleListX = new int[3];
		int[] triangleListY = new int[3];
		int startX, startY, orient;

		if (g == null)
			return;
		orient = (((i + j + ((invert) ? 0 : 1)) % MAX_ORIENT) +
			MAX_ORIENT) % MAX_ORIENT;
		for (int k = 0; k < 3; k++) {
			triangleListX[k] = (width / 2) *
				triangleUnit[orient][k].x +
				((k > 0) ? triangleListX[k - 1] : 0);
			triangleListY[k] = height *
				triangleUnit[orient][k].y +
				((k > 0) ? triangleListY[k - 1] : 0);
		}
		startX = x + width / 2 * (i + 1) + 1;
		startY = y + height * j + ((orient == 0) ? height : 0);
		g.setColor(cid);
		pm = pmid;
		fillPolygon(g, startX, startY, triangleListX, triangleListY,
			(orient == 0), 3, piece);
		g.setColor(bid);
		drawPolygon(g, startX, startY, triangleListX, triangleListY,
			3/*, piece*/);
	}

	void drawUnitBorder(Graphics g, boolean invert,
			int x, int y, int i, int j, int width, int height) {
		int[] triangleListX = new int[3];
		int[] triangleListY = new int[3];
		int startX, startY, orient;

		if (g == null)
			return;
		orient = (((i + j + ((invert) ? 0 : 1)) % MAX_ORIENT) +
			MAX_ORIENT) % MAX_ORIENT;
		for (int k = 0; k < 3; k++) {
			triangleListX[k] = (width / 2) *
				triangleUnit[orient][k].x +
				((k > 0) ? triangleListX[k - 1] : 0);
			triangleListY[k] = height *
				triangleUnit[orient][k].y +
				((k > 0) ? triangleListY[k - 1] : 0);
		}
		startX = x + width / 2 * (i + 1) + 1;
		startY = y + height * j + ((orient == 0) ? height : 0);
		drawPolygon(g, startX, startY, triangleListX, triangleListY,
			3/*, false*/);
	}

	void drawUnit(Graphics g, int pmid, int cid, Color bid, boolean invert,
			int x, int y, int i, int j, int width, int height,
			boolean piece) {
		drawUnit(g, pmid, colorThings[cid], bid, invert,
			x, y, i, j, width, height, piece);
	}

	static void clearUnits(Graphics g, Color color,
			 int x, int y, int i, int j, int width, int height) {
		if (g == null)
			return;
		g.setColor(color);
		g.fillRect(x + width * i, y + height * j, width, height);
	}

	void fillPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY,
			boolean inverted, int n, boolean piece) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		if (!piece) {
			g.fillPolygon(tempListX, tempListY, n);
			return;
		}
		if (!plainPiece) {
			int[] insideListX = new int[3];
			int[] insideListY = new int[3];
			int[] borderX = new int[4];
			int[] borderY = new int[4];
			int i, offset;
			int midPoint = (tempListY[1] - tempListY[0]) / 3;
			int EDGE = 2;

			g.fillPolygon(tempListX, tempListY, n);
			offset = (int) (midPoint / (SQRT3 * EDGE / 2.0));
			for (i = 0; i < 3; i++) {
				insideListX[i] = ((EDGE - 1) * tempListX[i] +
					tempListX[0]) / EDGE;
				insideListY[i] = ((EDGE - 1) * tempListY[i] +
					tempListY[0] +
					midPoint) / EDGE + offset;
			}
			Color c = g.getColor();
			for (i = 0; i < 3; i++) {
				if ((!inverted && (((pm & (1 << (2 - ((i + 1) % 3)))) != 0))) ||
						(inverted && (((pm & (1 << (2 - ((i + 2) % 3))))) != 0))) {

					borderX[0] = tempListX[i];
					borderY[0] = tempListY[i];
					borderX[1] = insideListX[i];
					borderY[1] = insideListY[i];
					borderX[2] = insideListX[(i + 1) % 3];
					borderY[2] = insideListY[(i + 1) % 3];
					borderX[3] = tempListX[(i + 1) % 3];
					borderY[3] = tempListY[(i + 1) % 3];
					g.setColor(c.darker());
					g.fillPolygon(borderX, borderY, 4);
					g.setColor(brighter(c));
					g.drawPolygon(borderX, borderY, 4);
				}
			}
			return;
		}
		Color c = g.getColor();
		int[] X = new int[n];
		int[] Y = new int[n];
		Point midPoint = new Point(tempListX[0],
			(tempListY[0] + tempListY[1] + tempListY[2]) / 3);

		X[0] = midPoint.x;
		Y[0] = midPoint.y;
		X[1] = tempListX[1];
		Y[1] = tempListY[1];
		X[2] = tempListX[2];
		Y[2] = tempListY[2];
		if (inverted)
			g.setColor(brighter(c));
		else
			g.setColor(c.darker());
		g.fillPolygon(X, Y, n);
		X[1] = tempListX[0];
		Y[1] = tempListY[0];
		X[2] = tempListX[1];
		Y[2] = tempListY[1];
		g.setColor(c);
		g.fillPolygon(X, Y, n);
		X[1] = tempListX[2];
		Y[1] = tempListY[2];
		X[2] = tempListX[0];
		Y[2] = tempListY[0];
		if (inverted)
			g.setColor(c.darker());
		else
			g.setColor(brighter(c));
		g.fillPolygon(X, Y, n);
	}

	static void drawPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n/*, boolean piece*/) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);

		/*if (!piece || plainPiece)
			return;

		g.setColor(this.background);
		Point midPoint = new Point(tempListX[0],
			(tempListY[0] + tempListY[1] + tempListY[2]) / 3);
		int[] innerListX = new int[n];
		int[] innerListY = new int[n];

		innerListX[0] = (tempListX[0] + midPoint.x) / 2;
		innerListY[0] = (tempListY[0] + midPoint.y) / 2;
		innerListX[1] = (tempListX[1] + midPoint.x) / 2;
		innerListY[1] = (tempListY[1] + midPoint.y) / 2;
		innerListX[2] = (tempListX[2] + midPoint.x) / 2;
		innerListY[2] = (tempListY[2] + midPoint.y) / 2;
		g.drawPolygon(innerListX, innerListY, n);*/
	}

	/* Pure colors can get brighter */
	static Color brighter(Color color) {
		int r = color.getRed();
		int g = color.getGreen();
		int b = color.getBlue();
		final double FACTOR = 0.8;

		/* From 2D group:
		 * 1. black.brighter() should return grey
		 * 2. applying brighter to blue will always return blue, brighter
		 * 3. non pure color (non zero rgb) will eventually return white
		 */
		int i = (int) ((1 - FACTOR) * 255);

		if (r >= 0 && r < i)
			r = i;
		if (g >= 0 && g < i)
			g = i;
		if (b >= 0 && b < i)
			b = i;
		return new Color(Math.min((int) (r / FACTOR), 255),
			Math.min((int) (g / FACTOR), 255),
			Math.min((int) (b / FACTOR), 255));
	}

	void setPlainPiece(boolean plainPiece) {
		this.plainPiece = plainPiece;
	}

	boolean getPlainPiece() {
		return this.plainPiece;
	}
}
