package com.sillycycle.bagleyd.tertris;

/*
 * @(#)TertrisCanvas.java
 *
 * Copyright 1993 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>TertrisCanvas</code> class does all the graphics for
 * <code>TertrisFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class TertrisCanvas extends Canvas {
	private static final long serialVersionUID = 42L;

	transient PolyMode[] tris; //[PolyiamondThing.MAX_UNITS - PolyiamondThing.MIN_UNITS + 1];

	Random generator;
	int startlevel = TertrisInterface.DEFAULT_LEVEL;
	int level = TertrisInterface.DEFAULT_LEVEL;
	int rows = TertrisInterface.DEFAULT_ROWS;
	int columns = TertrisInterface.DEFAULT_COLUMNS;
	int fill = TertrisInterface.DEFAULT_FILL;
	boolean fixedFill = false, suddenAppear = false;
	boolean bonus = false;
	boolean cw = true, grid = true, showNext = false;
	boolean invert = false, dropPaused = false;
	boolean firstPaint = false, resizePaint = false;
	boolean redrawPaint = false, framePaint = false;
	boolean focus = true;
	int score = 0, completedRows = 0;
	int triWidth = 38, triHeight = 31; /* triWidth*3^(1/2)/2 */
	int viewWidth = triWidth * (columns / 2 + 1) + 1;
	int viewHeight = triHeight * rows + 1;
	Point fieldOffset = new Point(2, 2);
	Color background;
	transient Thing curThing, nextThing;
	transient Spec curSpec = new Spec(PolyiamondThing.DEFAULT_UNITS);
	transient Spec nextSpec = new Spec(PolyiamondThing.DEFAULT_UNITS);
	PolyiamondImage polyiamondImage;
	transient Field[][] field = null;
	JFrame frame = null;
	static final double SQRT3_2 = 0.866054; /* sqrt(3) / 2 */
	boolean debug = false;

	public TertrisCanvas() {
		curThing = new Thing();
		nextThing = new Thing();
		polyiamondImage = new PolyiamondImage(Color.black);
		generator = new Random(System.nanoTime());
		readPolyiamonds();
		sizeInit();
		inits();
	}

	void specInit(int units, boolean diagonal, boolean mixed) {
		curSpec = new Spec(units, diagonal, mixed, false);
		nextSpec = new Spec(units, diagonal, mixed, false);
	}

	final void sizeInit() {
		field = new Field[rows][columns];
		for (int y = 0; y < rows; y++)
			for (int x = 0; x < columns; x++)
				field[y][x] = new Field();
	}

	final void inits() {
		int turnStyle, numberPolyiamonds;

		turnStyle = tris[nextSpec.units - PolyiamondThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].turnStyle;
		if (turnStyle < 0 || turnStyle >= PolyiamondThing.MAX_TURN_TYPES) {
			System.err.println("Tertris: corrupted polyiamond file");
			System.err.println("turnStyle = " + turnStyle);
		}
		numberPolyiamonds = tris[nextSpec.units - PolyiamondThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[(nextSpec.mixed) ? 1 : 0][turnStyle];
		if (numberPolyiamonds <= 0 ||
				numberPolyiamonds > PolyiamondThing.MAX_START_POLYIAMONDS) {
			System.err.println("Tertris: corrupted polyiamond file");
			System.err.println("\tnumberPolyiamonds = " + numberPolyiamonds);
		}
		initFill();
		/* new things */
		newThing();
		newThing();
	}

	void newThing() {
		polyiamondImage.setBackground(background);
		curThing.clone(nextThing);
		curSpec.clone(nextSpec);
		nextThing.setRandomNumber(Math.abs(generator.nextInt()));
		nextThing.setRotation(generator.nextInt(TertrisInterface.MAX_SIDES));
		nextThing.setReflection(generator.nextInt(2));
		nextThing.setBonus(nextSpec.bonus);
		redoNext();
	}

	void drawTitle() {
		//TertrisApp.drawTitle
	}

	void drawStatus() {
		/*
		TertrisApp.drawScore("Score: " + score);
		TertrisApp.drawLevel("Level: " + level);
		TertrisApp.drawRows("Rows: " + complededRows);
		*/
	}

	void resizeField() {
		triWidth = getSize().width;
		triHeight = getSize().height;
		if ((triHeight - 2) / rows < ((int) ((triWidth - 2) *
				SQRT3_2 / (columns / 2 + 1))))
			triHeight = (triHeight - 2) / rows;
		else
			triHeight = (int) ((triWidth - 2) * SQRT3_2 /
				(columns / 2 + 1));
		triWidth = (int) (triHeight / SQRT3_2);
		viewWidth = triWidth * (columns / 2 + 1) + 1;
		viewHeight = triHeight * rows + 1;
		fieldOffset = new Point((getSize().width - 2 - viewWidth) / 2,
			(getSize().height - 2 - viewHeight) / 2);
	}

	void clearUnit(Graphics g, int i, int j) {
		polyiamondImage.drawUnit(g, PolyiamondImage.NUM_BITMAPS - 1,
			background, ((grid) ? Color.black : background),
			invert, fieldOffset.x, fieldOffset.y,
			i, j, triWidth, triHeight, false);
	}

	void xorUnit(Graphics g, int i, int j, int k) {
		if (field[j][i].pmid >= 0) {
			Color color = PolyiamondImage.colorThings[field[j][i].cid];

			if (k % 2 == 0) // Simulate Xor
				color = new Color(255 - color.getRed(),
					255 - color.getGreen(),
					255 - color.getBlue());
			polyiamondImage.drawUnit(g, field[j][i].pmid,
				color, ((grid) ? Color.black : background),
				invert, fieldOffset.x, fieldOffset.y,
				i, j, triWidth, triHeight, true);
		}
	}

	void clearField() {
		Graphics g = this.getGraphics();

		// This will flash
		PolyiamondImage.clearUnits(g, Color.white, 0, 0, 0, 0,
			getSize().width, getSize().height);
		resizeField();
		for (int i = 0; i < columns; i++)
			for (int j = 0; j < rows; j++)
				clearUnit(g, i, j);
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = fieldOffset.x - 2;
		int offsetY = fieldOffset.y - 2;
		int lengthX = fieldOffset.x + viewWidth + 2;
		int lengthY = fieldOffset.y + viewHeight + 2;

		if (g == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, offsetY);
		g.fillRect(0, offsetY, offsetX, lengthY - offsetY);
		g.fillRect(0, lengthY, getSize().width, offsetY);
		g.fillRect(lengthX, offsetY,
			getSize().width - lengthX, lengthY - offsetY);
		draw3DFrame(show && focus, offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void drawField(int start, int finish) {
		Graphics g = this.getGraphics();
		int i, j;

		for (i = 0; i < columns; i++)
			for (j = start; j < finish; j++) {
				if (field[j][i].pmid >= 0)
				  polyiamondImage.drawUnit(g, field[j][i].pmid,
				    field[j][i].cid,
				    ((grid) ? Color.black : background),
				    invert, fieldOffset.x, fieldOffset.y,
				    i, j, triWidth, triHeight, true);
				else
				  clearUnit(g, i, j);
			}
	}

	void drawField() {
		drawField(0, rows);
	}

	void drawThing() {
		Graphics g = this.getGraphics();
		int i, j;

		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++) {
			int xi = curThing.position.x + i;

			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++) {
				int yj = curThing.position.y + j;

				if (yj >= 0 &&
				      PolyiamondImage.inHexagon(curThing.size, i, j) &&
				      (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0)) {
				    polyiamondImage.drawUnit(g,
				    PolyiamondImage.polyPixmap(i, j,
				    (tris[curThing.unitsIndex].poly
				    [curThing.polyNumber].shape)/*,
				    curThing.size*/),
				    curThing.colorNumber,
				    ((grid) ? Color.black : background),
				    invert, fieldOffset.x, fieldOffset.y,
				    xi, yj, triWidth, triHeight, true);
				}
			}
		}
	}

	void drawThingDiff(Thing oldThing) {
		Graphics g = this.getGraphics();
		int i, j;
		int sx = PolyiamondImage.hexCol(curThing.size);
		int sy = PolyiamondImage.hexRow(curThing.size);

		for (i = 0; i < sx; i++) {
			int xi = curThing.position.x + i;

			for (j = 0; j < sy; j++) {
				int yj = curThing.position.y + j;

				if (yj >= 0 &&
				      PolyiamondImage.inHexagon(curThing.size, i, j) &&
				      (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0))
				  polyiamondImage.drawUnit(g,
				    PolyiamondImage.polyPixmap(i, j,
				    (tris[curThing.unitsIndex].poly
				    [curThing.polyNumber].shape)/*,
				    curThing.size*/),
				    curThing.colorNumber,
				    ((grid) ? Color.black : background),
				    invert, fieldOffset.x, fieldOffset.y,
				    xi, yj, triWidth, triHeight, true);
			}
		}
		for (i = 0; i < sx; i++) {
			int xi = oldThing.position.x + i;
			int ox = xi - curThing.position.x;

			if (xi >= 0 && xi < columns)
				for (j = 0; j < sy; j++) {
					int yj = oldThing.position.y + j;
					int oy = yj - curThing.position.y;

					if (yj >= 0 && yj < rows &&
						(tris[oldThing.unitsIndex].poly
						[oldThing.polyNumber].shape[j][i] != 0) &&
					    ((ox < 0) || (ox >= sx) ||
					     (oy < 0) || (oy >= sy) ||
					     tris[curThing.unitsIndex].poly
					      [curThing.polyNumber].shape[oy][ox] == 0))
					    clearUnit(g, xi, yj);
				}
		}
	}

	void drawNext(JFrame app) {
		((TertrisFrame) app).drawNext(nextThing, tris);
	}

	void clearNext(JFrame app) {
		((TertrisFrame) app).clearNext(nextThing, tris);
	}

	void redoNext() {
		int diagonal = (nextSpec.diagonal) ? 1 : 0;
		int turnStyle =
			tris[nextSpec.units - PolyiamondThing.MIN_UNITS].mode
				[diagonal].turnStyle;
		int nextStart, i, diffPolys;

		/* thing units are offset */
		nextThing.unitsIndex = nextSpec.units - PolyiamondThing.MIN_UNITS +
			((nextThing.getBonus()) ? 1 : 0);
		diffPolys = tris[nextThing.unitsIndex].mode[diagonal].number
			[(nextThing.getBonus()) ? PolyiamondThing.NOMIX :
			((nextSpec.mixed) ? PolyiamondThing.MIXED :
			PolyiamondThing.NOMIX)][turnStyle];
		if (diffPolys == 0) {
			System.out.println("no poly units=" + nextThing.unitsIndex + PolyiamondThing.MIN_UNITS
				+ ", diagonal=" + nextSpec.diagonal
				+ ", mixed=" + nextSpec.mixed
				+ ", bonus=" + nextThing.getBonus());
			return;
		}
		nextStart = nextThing.getRandomNumber() % diffPolys;
		nextThing.colorNumber = nextStart;
		if (nextSpec.mixed && !nextThing.getBonus()) {
			nextThing.unitsIndex = 0;
			while (nextStart >= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyiamondThing.NOMIX][turnStyle]) {
				nextStart -= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyiamondThing.NOMIX][turnStyle];
				nextThing.unitsIndex++;
				if (nextThing.unitsIndex > nextSpec.units - PolyiamondThing.MIN_UNITS) {
					System.out.println("Something wrong");
				}
			}
		}
		nextThing.polyNumber = tris[nextThing.unitsIndex].mode
			[diagonal].start[nextStart][turnStyle];
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle >
				PolyiamondThing.NONE)
			for (i = 0; i < nextThing.getRotation(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].rotation;
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle ==
				PolyiamondThing.ALL)
			for (i = 0; i < nextThing.getReflection(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].reflection;
		nextThing.size = tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].size;
		/* nextThing.position.x = generator.nextInt(columns - nextThing.size + 1); */
		nextThing.position.x = (columns - nextThing.size) / 2;
		nextThing.position.y = -tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].startHeight[(suddenAppear) ? 1 : 0];
		if (((nextThing.position.x - nextThing.position.y +
				((invert) ? 0 : 1) + nextThing.size) % 2) != 0)
			nextThing.position.x--;
		nextThing.direction = ((generator.nextInt(2) == 1) ? 1 : -1);
		if (debug)
			System.out.println("Size: " + nextThing.size +
				", Poly: " + nextThing.polyNumber +
				", Color: " + nextThing.colorNumber);
	}

	boolean getPlainPiece() {
		return polyiamondImage.getPlainPiece();
	}

	void setPlainPiece(boolean plainPiece) {
		polyiamondImage.setPlainPiece(plainPiece);
		redrawPaint = true;
	}

	void banner(String msg) {
		((TertrisFrame) frame).displayMessage(msg);
	}

	void putUnit() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++)
			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++) {
				xi = x + i;
				yj = y + j;
				if ((yj >= 0) &&
				    (tris[curThing.unitsIndex].poly
				     [curThing.polyNumber].shape[j][i] != 0)) {
					field[yj][xi].pmid = PolyiamondImage.NUM_BITMAPS -
					   tris[curThing.unitsIndex].poly
					  [curThing.polyNumber].shape[j][i] - 1;
					field[yj][xi].cid = curThing.colorNumber;
				}
			}
	}

	boolean overlapping() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		if (field == null)
			return false;
		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++)
			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++)
			  if (tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    xi = x + i;
			    yj = y + j;
			    if ((yj >= rows) || (xi < 0) || (xi >= columns))
			      return true;
			    if (suddenAppear) { /* creationalism */
/* This method does not allow turning polyiamond to an area above screen */
			      if ((yj < 0) || (field != null && field != null && field[yj][xi].pmid >= 0))
				return true;
			    } else { /* evolution */
/* This method one can turn polyiamond to an area above of screen, also
   part of the polyiamond may not be visible initially */
			      if ((yj >= 0) && (field != null && field != null && field[yj][xi].pmid >= 0))
				return true;
			    }
			  }
		return false;
	}

	boolean overlapped(int dir,  boolean fall) {
		int i, j, oldxi, oldyj;
		int oldx = curThing.position.x - ((fall) ? 1 : 2) * dir;
		int oldy = curThing.position.y - ((fall) ? 1 : 0);

		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++)
			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++)
			  if ((curThing.position.y + j >= 0) && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    oldxi = oldx + i;
			    oldyj = oldy + j;
			    if (((oldxi + oldyj + ((invert) ? 1 : 0)) & 1) == 1) {
			      if ((oldyj >= 0) && (field != null && field[oldyj][oldxi + dir].pmid >= 0))
				return true;
			    } else {
			      if (field != null && field[oldyj + ((fall) ? 1 : 0)][oldxi + ((fall) ? 0 : dir)].pmid >= 0)
				return true;
			    }
			  }
		return false;
	}

	boolean atBottom() {
		if (collision(curThing.direction))
			return collision(-curThing.direction);
		return false;
	}

	boolean collision(int dir) {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++)
			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++) {
			  xi = x + i;
			  yj = y + j;
			  if ((yj >= -1) && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0 &&
			      ((yj >= rows - 1) ||
			       (xi + dir < 0) || (xi + dir >= columns) ||
			       (field != null && field[yj + 1][xi + dir].pmid >= 0)))
			     return true;
			}
		for (i = 0; i < PolyiamondImage.hexCol(curThing.size); i++)
			for (j = 0; j < PolyiamondImage.hexRow(curThing.size); j++) {
			  xi = x + i;
			  yj = y + j;
			  if ((yj >= -1) && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    if (((xi + yj + ((invert) ? 1 : 0)) & 1) == 1) {
			      if ((yj >= 0) &&
				 (field != null && field[yj][xi + dir].pmid >= 0))
				return true;
			    } else {
			      if (field != null && field[yj + 1][xi].pmid >= 0)
				return true;
			    }
			  }
			}
		return false;
	}

	void tryMove(int move) {
		Thing oldThing = new Thing();
		Thing midThing = new Thing();
		int i;

		oldThing.clone(curThing);
		switch (move) {
		case TertrisInterface.FALL:
			curThing.position.y++;
			curThing.position.x += curThing.direction;
			if (overlapping() ||
			    overlapped(curThing.direction, true)) {
				curThing.direction = -curThing.direction;
				curThing.position.x = oldThing.position.x +
					curThing.direction;
				if (overlapping() ||
				    overlapped(curThing.direction, true)) {
					curThing.clone(oldThing);
				} else {
					drawThingDiff(oldThing);
				}
			} else {
				drawThingDiff(oldThing);
			}
			break;
		case TertrisInterface.DROP:
			do {
				do {
					midThing.clone(curThing);
					curThing.position.y++;
					curThing.position.x += curThing.direction;
					if (!dropPaused)
						score += level + fill + ((grid) ? 0 : 1);
				} while (!overlapping() &&
					!overlapped(curThing.direction, true));
				curThing.direction = -curThing.direction;
				curThing.position.x = midThing.position.x +
					curThing.direction;
			} while (!overlapping() &&
				!overlapped(curThing.direction, true));
			curThing.clone(midThing);
			drawThingDiff(oldThing);
			break;

		case TertrisInterface.MOVE_LEFT:
			curThing.position.x -= 2;
			if (overlapping() || overlapped(-1, false))
				curThing.clone(oldThing);
			else
				drawThingDiff(oldThing);
			break;

		case TertrisInterface.MOVE_RIGHT:
			curThing.position.x += 2;
			if (overlapping() || overlapped(1, false))
				curThing.clone(oldThing);
			else
				drawThingDiff(oldThing);
			break;

		case TertrisInterface.ROTATE:
			if (tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
				[(curSpec.diagonal) ? 1 : 0].turnStyle > PolyiamondThing.NONE) {
			  if (cw)
			    for (i = 0; i < TertrisInterface.MAX_SIDES - 1; i++)
			      curThing.polyNumber = tris[curThing.unitsIndex].poly
				[curThing.polyNumber].rotation;
			  else /* ccw */
			    curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].rotation;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			}
			if (overlapping())
			  curThing.clone(oldThing);
			else
			  drawThingDiff(oldThing);
			break;
		case TertrisInterface.REFLECT: /* reflect on y axis */
			if (tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
			    [(curSpec.diagonal) ? 1 : 0].turnStyle == PolyiamondThing.ALL) {
			  curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].reflection;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			}
			if (overlapping())
			  curThing.clone(oldThing);
			else
			  drawThingDiff(oldThing);
			break;
		default:
		}
		if (redrawPaint) {
			drawField();
			redrawPaint = false;
		}
		if (framePaint) {
			drawFrame(focus);
			framePaint = false;
		}
		if (overlapping())
			curThing = oldThing;
		else
			drawThingDiff(oldThing);
	}

	void initFill() {
		int i, j, col, fixedUnits;
		int turnStyle = tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int mixed = (nextSpec.mixed && TertrisInterface.mixedLegal(nextSpec.units, nextSpec.mixed)) ? 1 : 0;
		int numberPolyiamonds = tris[nextSpec.units - PolyiamondThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[mixed][turnStyle];

		if (fixedFill) {
			/* This algorithm randomly picks fixedUnits per row
			 * (4 for DEFAULTWIDTH, ie. 10/2-1)
			 * This is more in keeping with the original */
			fixedUnits = columns / 2 - 1;
			for (j = 0; j < rows; j++) {
				for (i = 0; i < columns; i++) {
					field[j][i].pmid = -1;
					field[j][i].cid = 0;
				}
				if (j >= rows - fill)
					for (i = 0; i < fixedUnits; i++) {
						do
							col = generator.nextInt(columns);
						while (field[j][col].pmid != -1);
						field[j][col].pmid = PolyiamondImage.NUM_BITMAPS - 1;
						field[j][col].cid = generator.nextInt(numberPolyiamonds);
					}
			}
		} else {
			/* This algorithm randomly fills triangles to a certain level
			   The problem with this is that it could potentially fill
			   up a whole row */
			for (i = 0; i < columns; i++)
				for (j = 0; j < rows; j++)
					if ((j >= rows - fill) &&
							(generator.nextInt(2) == 0)) {
						field[j][i].pmid = PolyiamondImage.NUM_BITMAPS - 1;
						field[j][i].cid = generator.nextInt(numberPolyiamonds);
					} else {
						field[j][i].pmid = -1;
						field[j][i].cid = 0;
					}
		}
	}

	void fillLines() {
		Graphics g = this.getGraphics();
		int i, j;
		int turnStyle = tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int pmid = PolyiamondImage.NUM_BITMAPS - 1;

		for (j = 0; j <= rows / 2; j++) {
			for (i = 0; i < columns; i++) {
				polyiamondImage.drawUnit(g,
				  pmid,
				  generator.nextInt(tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
				  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
				  [turnStyle]),
				  ((grid) ? Color.black : background),
				  invert, fieldOffset.x, fieldOffset.y,
				  i, j, triWidth, triHeight, true);
				polyiamondImage.drawUnit(g,
				  pmid,
				  generator.nextInt(tris[curSpec.units - PolyiamondThing.MIN_UNITS].mode
				  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
				  [turnStyle]),
				  ((grid) ? Color.black : background),
				  invert, fieldOffset.x, fieldOffset.y,
				  i, rows - j - 1, triWidth, triHeight, true);
			}
		}
	}

	int checkLines() {
		Graphics g = this.getGraphics();
		int[] lSet = new int[rows];
		int nSet = 0, i, j, y;

		/* Find filled rows */
		for (j = 0; j < rows; j++) {
			lSet[j] = 0;
			for (i = 0; i < columns; i++)
				if (field[j][i].pmid >= 0)
					lSet[j]++;
			if (lSet[j] == columns)
				nSet++;
		}
		if (nSet == 0) {
			return 0;
		}
		for (int k = 0; k < 2 * TertrisInterface.NUM_FLASHES; k++) {
		  for (j = 0; j < rows; j++) {
		    if (lSet[j] == columns)
		      for (i = 0; i < columns; i++) {
			xorUnit(g, i, j, k);
		      }
		  }
		  try {
		    Thread.sleep(100L);
		  } catch (InterruptedException e) {
		    e.printStackTrace();
		  }
		}
		/* Shrink */
		for (j = rows - 1; j >= 0; j--)
			if (lSet[j] == columns) {
			  for (y = j; y > 0; y--)
			    for (i = 0; i < columns; i++) {
			      field[y][i].pmid = field[y - 1][i].pmid;
			      field[y][i].cid = field[y - 1][i].cid;
			    }
			  for (i = 0; i < columns; i++)
			    field[0][i].pmid = -1;
			  drawField(0, j + 1);

			  for (i = j; i > 0; i--)
			    lSet[i] = lSet[i - 1];
			  lSet[0] = 0;
			  if (j > 0)
			    for (i = 0; i < columns; i++) {
			      int tmp = field[j][i].pmid;

			      if ((((i + j + ((invert) ? 0 : 1)) & 1) == 0) &&
				  (tmp >= 0) && (tmp != PolyiamondImage.checkDown(tmp))) {
				field[j][i].pmid = PolyiamondImage.checkDown(tmp);
				polyiamondImage.drawUnit(g,
				  field[j][i].pmid, field[j][i].cid,
				  ((grid) ? Color.black : background), invert,
				  0, 0, i, j, triWidth, triHeight, false);
			      }
			    }
			  j++;
			  for (int k = j; k < rows; k++) {
			    for (i = 0; i < columns; i++) {
			      int tmp = field[k][i].pmid;

			      if (tmp >= 0) {
				if (((i + k + ((invert) ? 0 : 1)) & 1) == 1)
				  tmp = PolyiamondImage.checkDown(tmp);
				if (((i + k + ((invert) ? 0 : 1)) & 1) == 0)
				  tmp = PolyiamondImage.checkUp(tmp);
				field[k][i].pmid = PolyiamondImage.swapUpDown(tmp);
			      }
			    }
			  }
			  if (((TertrisFrame) frame).getToggleSound())
			     ((TertrisFrame) frame).playBumpAudio();
			}
		if ((nSet & 1) == 1) {
			invertGrid();
		}
		clearField();
		drawField();
		return nSet;
	}

	void invertGrid() {
		invert = !invert;
		nextThing.position.x = (columns - nextThing.unitsIndex) / 2;
		if ((nextThing.position.x - nextThing.position.y +
				((invert) ? 0 : 1) + nextThing.size) % 2 != 0)
			nextThing.position.x++;
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;

		if (g == null)
			return;
		if (grid)
			color = Color.black;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < rows; j++)
			for (int i = 0; i < columns; i++)
				polyiamondImage.drawUnitBorder(g,
					invert,
					fieldOffset.x, fieldOffset.y,
					i, j, triWidth, triHeight);
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			firstPaint = false;
			resizePaint = true;
		}
		if (resizePaint) {
			clearField();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(g, focus);
			framePaint = false;
		}
		redrawPaint = false;
		drawField();
		drawThing();
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	PolyiamondThing polyiamondThing = null;
	PolyiamondThingFallback polyiamondThingFallback = null;
	final void readPolyiamonds() {
		polyiamondThing = new PolyiamondThing();
		tris = PolyiamondThing.readObjectsXML(TertrisInterface.XML_FILE, null);
		if (tris == null) {
			System.err.println("Using fallback pieces.");
			polyiamondThingFallback = new PolyiamondThingFallback();
			tris = polyiamondThingFallback.readStatic();
		}
		/*if (debug) {
			checkPolyiamonds();
		}*/
	}
}

class Spec {
	volatile int units;
	volatile boolean diagonal, mixed, bonus;

	public Spec(int units,
			boolean diagonal, boolean mixed, boolean bonus) {
		this.units = units;
		this.diagonal = diagonal;
		this.mixed = mixed;
		this.bonus = bonus;
	}

	public Spec(int units) {
		this.units = units;
		this.diagonal = this.mixed = this.bonus = false;
	}

	void clone(Spec nextSpec) {
		units = nextSpec.units;
		diagonal = nextSpec.diagonal;
		mixed = nextSpec.mixed;
		bonus = nextSpec.bonus;
	}

	public void setDiagonal(boolean diagonal) {
		this.diagonal = diagonal;
	}

	public void setMixed(boolean mixed) {
		this.mixed = mixed;
	}

	public void setBonus(boolean bonus) {
		this.bonus = bonus;
	}

	public void setUnits(int units) {
		this.units = units;
	}
}

class Field {
	int pmid, cid; // pixmaps and colors

	public Field() {
		pmid = -1;
		cid = 0;
	}
}
