package com.sillycycle.bagleyd.tertris;

/*
 * @(#)TertrisFrame.java
 *
 * Copyright 1994 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;
import com.sillycycle.bagleyd.util.VerticalBorderLayout;

/**
 * The <code>TertrisFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class TertrisFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final String SCOREFILE = "tertris.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Pause %", "Clear &", "Setup ~",
		"Description ?", "Features !", "References ^", "About",
		"Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'%', '&', '~',
		'?', '!', '^', 'a', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 310;
	int heightValue = 635;
	static int xValue = 0;
	static int yValue = 0;
	Color bgColor = Color.gray.brighter();
	boolean sound = false;
	boolean controlValue = true;
	int levelValue = TertrisInterface.DEFAULT_LEVEL;
	int maxScore = -1;
	boolean debug = false;
	Next nextCanvas = new Next(PolyiamondThing.HEX_SIZE,
		TertrisInterface.NEXTTRIWIDTH, TertrisInterface.NEXTTRIHEIGHT,
		bgColor);
	Tertris game = null;
	JPanel leftPanel = new JPanel();
	JTextField rowsTextField = new JTextField(10);
	JTextField scoreTextField = new JTextField(10);
	JTextField maxScoreTextField = new JTextField(10);
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	TertrisDialog gameDialog = null;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner levelChanger = null;
	//Slider levelChanger = null;

	public String getGame() {
		return this.stringBuilderSave.toString();
	}

	/*public void setGame(String s) {
		game.stringSave = s;
	}*/

	public void startGame() {
		game.start();
	}

	public void endGame() {
		game.gameOver(true);
	}

	public void stopGame() {
		game.pause();
	}

	public void clearGame() {
		game.gameOver(true);
		game.start();
	}

	public boolean processKey(char key) {
		return game.processKey(key);
	}

	public synchronized void setLevel(int i) {
		this.levelValue = i;
		if (levelChanger != null)
			levelChanger.setValue(i);
	}

	public void setUnits(int i) {
		gameDialog.setUnits(i);
		if (game != null)
			game.setUnits(i);
	}

	public static int getUnits() {
		return TertrisDialog.getUnits();
	}

	public void setColumns(int i) {
		gameDialog.setColumns(i);
		if (game != null)
			game.setColumns(i);
	}

	public static int getColumns() {
		return TertrisDialog.getColumns();
	}

	public void setFill(int i) {
		gameDialog.setFill(i);
		if (game != null)
			game.setFill(i);
	}

	public static int getFill() {
		return TertrisDialog.getFill();
	}

	public void setFixedFill(boolean flag) {
		gameDialog.setFixedFill(flag);
		if (game != null)
			game.setFixedFill(flag);
	}

	public static boolean getFixedFill() {
		return TertrisDialog.getFixedFill();
	}

	public void setSuddenAppear(boolean flag) {
		gameDialog.setSuddenAppear(flag);
		if (game != null)
			game.setSuddenAppear(flag);
	}

	public static boolean getSuddenAppear() {
		return TertrisDialog.getSuddenAppear();
	}

	public void setDiagonal(boolean flag) {
		gameDialog.setDiagonal(flag);
		if (game != null)
			game.setDiagonal(flag);
	}

	public static boolean getDiagonal() {
		return TertrisDialog.getDiagonal();
	}

	public void setMixed(boolean flag) {
		gameDialog.setMixed(flag);
		if (game != null)
			game.setMixed(flag);
	}

	public static boolean getMixed() {
		return TertrisDialog.getMixed();
	}

	public void setBonus(boolean flag) {
		gameDialog.setBonus(flag);
		if (game != null)
			game.setBonus(flag);
	}

	public static boolean getBonus() {
		return TertrisDialog.getBonus();
	}

	public void setClockwise(boolean flag) {
		gameDialog.setClockwise(flag);
		if (game != null)
			game.setClockwise(flag);
	}

	public static boolean getClockwise() {
		return TertrisDialog.getClockwise();
	}

	public void setPlainPiece(boolean flag) {
		gameDialog.setPlainPiece(flag);
		if (game != null)
			game.setPlainPiece(flag);
	}

	public static boolean getPlainPiece() {
		return TertrisDialog.getPlainPiece();
	}

	public void setShowNext(boolean flag) {
		gameDialog.setShowNext(flag);
		if (game != null)
			game.setShowNext(flag);
	}

	public boolean getShowNext() {
		return game.showNext;
	}

	public void setGrid(boolean flag) {
		gameDialog.setGrid(flag);
		if (game != null)
			game.setGrid(flag);
	}

	public static boolean getGrid() {
		return TertrisDialog.getGrid();
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		int unitsValue, columnsValue, fillValue;
		boolean diagonalValue, mixedValue;

		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		bgColor = getOpt("bgColor", bgColor);
		controlValue = getOpt("control", controlValue);
		levelValue = getOpt("level", levelValue);
		if (levelValue < TertrisInterface.MIN_LEVELS)
			levelValue = TertrisInterface.MIN_LEVELS;
		else if (levelValue > TertrisInterface.MAX_LEVELS)
			levelValue = TertrisInterface.MAX_LEVELS;
		unitsValue = getOpt("units", TertrisDialog.getUnits());
		if (unitsValue > PolyiamondThing.MAX_UNITS)
			unitsValue = PolyiamondThing.MAX_UNITS;
		else if (unitsValue < PolyiamondThing.MIN_UNITS)
			unitsValue = PolyiamondThing.DEFAULT_UNITS;
		diagonalValue = getOpt("diagonal", TertrisDialog.getDiagonal());
		mixedValue = getOpt("mixed", TertrisDialog.getMixed());
		gameDialog.setUnits(unitsValue);
		gameDialog.setDiagonal(diagonalValue);
		gameDialog.setMixed(mixedValue);
		columnsValue = getOpt("columns", TertrisDialog.getColumns());
		if (columnsValue < TertrisInterface.MIN_COLUMNS)
			columnsValue = TertrisInterface.MIN_COLUMNS;
		if (columnsValue % 2 == 0)
			columnsValue += 1;
		gameDialog.setColumns(columnsValue);
		fillValue = getOpt("fill", TertrisDialog.getFill());
		if (fillValue < 0)
			fillValue = 0;
		else if (fillValue > TertrisInterface.MAX_FILL)
			fillValue = TertrisInterface.MAX_FILL;
		gameDialog.setFill(fillValue);
		gameDialog.setFixedFill(getOpt("fixedFill",
			TertrisDialog.getFixedFill()));
		gameDialog.setSuddenAppear(getOpt("suddenAppear",
			TertrisDialog.getSuddenAppear()));
		gameDialog.setBonus(getOpt("bonus",
			TertrisDialog.getBonus()));
		gameDialog.setClockwise(getOpt("cw",
			TertrisDialog.getClockwise()));
		gameDialog.setPlainPiece(getOpt("plain",
			TertrisDialog.getPlainPiece()));
		gameDialog.setShowNext(getOpt("showNext",
			TertrisDialog.getShowNext()));
		gameDialog.setGrid(getOpt("grid",
			TertrisDialog.getGrid()));
		gameDialog.setToggleSound(getOpt("sound",
			TertrisDialog.getToggleSound()));
	}

	public void setupGame() {
		if (aDialog == null || !OKDialog.getActive()) {
			gameDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Tertris", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"The ever popular falling tetromino game now with triangles.  " +
"Enhanced to allow all polyiamond classes that do not have" + NL +
"pieces with holes.  Mixing polyiamond classes (so as to " +
"permit polyiamonds made up by a variable number triangles)" + NL +
"can also be selected, as well as, 1 set of diagonal " +
"polyiamonds that do not have pieces with holes." + NL +
" " + NL +
"To make these games interesting and \"do-able\" the ability " +
"to rotate gets turned off if it is easier than the 5" + NL +
"triangle polyiamond.  An added ability \"reflect\" " +
"(right-left -> left-right) allows one to attempt the 6" + NL +
"polyiamond pieces." + NL +
" " + NL +
"To get high scores, do not use \"show next\" mode, try " +
"starting at a higher level and try to let \"tertris\" eat more" + NL +
"than one line at a time.  Polyiamonds with diagonal or " +
"with a higher number of triangles earn more points.  Mixed" + NL +
"mode earns less points.  Play fast and see how well you " +
"can do in the bonus mode.  Bonuses are set when the" + NL +
"level is changed automatically.  (Bonus mode not available " +
"for 8 triangles)." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Click \"mouse-left\" button , or press \"J\", \"j\", " +
"\"S\", or \"s\" keys to move the falling polyiamond left." + NL +
" " + NL +
"Click \"mouse-center\" button, or press \"K\", \"k\", " +
"\"D\", or \"d\" keys to rotate the falling polyiamond." + NL +
"Note this may be turned off if the depending on the " +
"\"turn-style\" but normally this is \"on\"." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"L\", \"l\", " +
"\"F\", or \"f\" keys to move the falling polyiamond right." + NL +
" " + NL +
"Click \"mouse-scroll-up\", or press \"I\", \"i\", " +
"\"E\", or \"e\" keys to flip (reflect) the falling polyiamond." + NL +
"Note this may be turned off if the depending on the " +
"\"turn-style\" but normally this is \"off\"." + NL +
" " + NL +
"Click \"mouse-scroll-down\", or press \"space-bar\" " +
"to drop the falling polyiamond." + NL +
" " + NL +
"The key pad and arrow keys are defined similar." + NL +
"Key pad is defined for Tertris as:" + NL +
"  /    Reflect" + NL +
"  8    Rotate" + NL +
"  ^" + NL +
"4<5>6  Left, Rotate, Right" + NL +
"  v" + NL +
"  2    Drop" + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up " +
"the falling polyiamond." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow down " +
"the falling polyiamond." + NL +
" " + NL +
"Press \"+\" or \"=\" keys to increase the " +
"triangles in the falling polyiamond.  Note: to keep the" + NL +
"level of play similar, the \"turn style\" can change.  " +
"The ability to reflect polyiamonds is available" + NL +
"only when the number of triangles >= 5.  The ability " +
"to rotate polyiamonds, is available only when" + NL +
"the number of triangles is >= 4.  Bonus polyiamonds " +
"still have the current \"turn style\"." + NL +
" " + NL +
"Press \"-\" or \"_\" keys to decrease the triangles in " +
"the falling polyiamond." + NL +
" " + NL +
"Press \"]\" or \"}\" keys to increase the column/row " +
"size." + NL +
" " + NL +
"Press \"[\" or \"{\" keys to decrease the column/row " +
"size." + NL +
" " + NL +
"Press \"`\", \"1\", \"2\", \"3\", \"4\", \"5\", " +
"\"6\", \"7\", \"8\", \"9\", or \"0\" keys" + NL +
"(not the keypad 0-9) to change the number of fill lines.  " +
"\"`\" for 0 lines and \"0\" for 10 lines." + NL +
" " + NL +
"Press \"X\" or \"x\" to toggle between fixed fill or " +
"random fill." + NL +
" " + NL +
"Press \"R\" or \"r\" to toggle between sudden initial " +
"appearance or gradual initial appearance." + NL +
" " + NL +
"Press \"G\" or \"g\" to toggle diagonal mode for 2 " +
"triangles." + NL +
" " + NL +
"Press \"M\" or \"m\" to toggle mixed mode.  This is " +
"ignored for 2 triangles." + NL +
" " + NL +
"Press \"B\" or \"b\" to toggle bonus mode.  This is " +
"ignored for 8 triangles." + NL +
" " + NL +
"Press \"W\" or \"w\" to toggle clockwise or counter " +
"clockwise turning." + NL +
" " + NL +
"Press \"P\" or \"p\" to toggle plain pieces or " +
"three dimensional pieces." + NL +
" " + NL +
"Press \"N\" or \"n\" to toggle show next." + NL +
" " + NL +
"Press \"#\" key to toggle the grid." + NL +
" " + NL +
"Press \"|\" key to pause." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide and pause program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL,
/* "The tertris may be resized.  Triangles will reshape depending on" + NL +
"the room they have." + NL + */
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"Based at one time on Q. Alex Zhao's Tetris." + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java TertrisFrame Version 8.2.7" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/altris.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = TertrisFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = TertrisFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playBumpAudio() {
		playClip(TertrisInterface.SOUNDS[TertrisInterface.Sounds.BUMP.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (game == null)
			return;
		game.clearField();
		game.setBounds(0, 0, width, height);
		game.drawField();
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width - leftPanel.getSize().width,
			secondaryPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		return String.valueOf(maxScore);
	}

	public synchronized void writeScore(int score) {
		if (score < maxScore)
			return;
		maxScore = score;
		maxScoreTextField.setText(String.valueOf(score));
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(String.valueOf(score));
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(int score) {
		maxScoreTextField.setText(String.valueOf(score));
	}

	public void initRecords(String cookieValue, boolean store) {
		if (cookieValue == null) { // if no cookie, then make one...
			if (store)
				writeScore(0);
		} else {
			if (debug)
				System.out.println("initRecords: " +
					cookieValue);
			maxScore = Integer.parseInt(cookieValue);
			maxScoreTextField.setText(String.valueOf(maxScore));
		}
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString().trim(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	public void displayMessage(String string) {
		if (frame != null) {
			if (string == null)
				frame.setTitle(TertrisInterface.TITLE);
			else
				frame.setTitle(TertrisInterface.TITLE +
					": " + string);
		}
	}

	@Override
	public void requestFocus() {
		if (game != null)
			game.requestFocus();
	}

	public void changeShowNext() {
		nextCanvas.setPlainPiece(game.getPlainPiece(),
			TertrisDialog.getShowNext());
	}

	public void drawNext(Thing nextThing, PolyMode[] polyMode) {
		if ((nextThing == null) || (polyMode == null))
			return;
		nextCanvas.setNextThingCanvas(nextThing, polyMode);
		nextCanvas.drawNext();
	}

	public void clearNext(Thing nextThing, PolyMode[] polyMode) {
		if ((nextThing == null) || (polyMode == null))
			return;
		nextCanvas.clearNext();
	}

	public void refreshNext() {
		game.redoNext();
		if (TertrisDialog.getShowNext()) {
			nextCanvas.clearNext();
			nextCanvas.drawNext();
		}
	}

	public void updateScore(int rows, int score) {
		if (rowsTextField == null || scoreTextField == null)
			return;
		rowsTextField.setText(String.valueOf(rows));
		scoreTextField.setText(String.valueOf(score));
	}

	public void initScore() {
		if (game != null && maxScore == -1) {
			readScoreRecords();
			maxScoreTextField.setText(String.valueOf(maxScore));
		}
	}

	public Image getImage(String fileName) {
		return Toolkit.getDefaultToolkit().getImage(fileName);
	}

	class LevelChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = levelValue;

			levelValue = levelChanger.getValue();
			if (oldValue != levelValue) {
				game.setLevel(levelValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public TertrisFrame(String title) {
		JPanel rowPanel = new JPanel();
		JPanel scorePanel = new JPanel();
		JPanel maxScorePanel = new JPanel();
		JPanel nextCanvasPanel = new JPanel();
		JPanel canvasPanel = new JPanel();

		try {
			mediumIcon = getIcon(TertrisInterface.ICONS[TertrisInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(TertrisInterface.ICONS[TertrisInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		readScoreRecords();
		gameDialog = new TertrisDialog(this,
			ComponentUtil.findJFrame(this),
			TertrisInterface.TITLE + ": Setup", "DONE");
		getOpts();
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(bgColor);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Pause %");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear &");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Setup ~");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			levelChanger = new Spinner("Level",
				levelValue, TertrisInterface.MIN_LEVELS,
				TertrisInterface.MAX_LEVELS);
			/*levelChanger = new Slider("Level",
				levelValue, TertrisInterface.MIN_LEVELS,
				TertrisInterface.MAX_LEVELS);*/
			rowPanel.add(new JLabel("Rows:"));
			rowPanel.add(rowsTextField);
			scorePanel.add(new JLabel("Score:"));
			scorePanel.add(scoreTextField);
			maxScorePanel.add(new JLabel("Max Score:"));
			maxScorePanel.add(maxScoreTextField);
			nextCanvasPanel.add(nextCanvas);
			leftPanel.setLayout(new ColumnLayout(5, 5, 5,
				ColumnLayout.RIGHT));
			leftPanel.add(levelChanger);
			leftPanel.add(rowPanel);
			leftPanel.add(scorePanel);
			leftPanel.add(maxScorePanel);
			leftPanel.add(nextCanvasPanel);
			secondaryPanel.setLayout(new VerticalBorderLayout());
			secondaryPanel.add(leftPanel, VerticalBorderLayout.LINE_START);
		}
		game = new Tertris(this, levelValue,
			TertrisDialog.getColumns(),
			TertrisDialog.getFill(),
			TertrisDialog.getFixedFill(),
			TertrisDialog.getSuddenAppear(),
			TertrisDialog.getBonus(),
			TertrisDialog.getClockwise(),
			TertrisDialog.getGrid(),
			TertrisDialog.getShowNext(),
			bgColor);
		canvasPanel.add(game);
		secondaryPanel.add(canvasPanel, VerticalBorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			levelChanger.addChangeListener(new LevelChangeListener());
			nextCanvas.setSize((PolyiamondImage.hexCol(
				PolyiamondThing.HEX_SIZE)) *
				TertrisInterface.NEXTTRIWIDTH / 2 + 4,
				PolyiamondImage.hexRow(PolyiamondThing.HEX_SIZE) *
				TertrisInterface.NEXTTRIHEIGHT + 4);
			game.setPlainPiece(TertrisDialog.getPlainPiece());
			nextCanvas.setPlainPiece(TertrisDialog.getPlainPiece(),
				TertrisDialog.getShowNext());
		}
		game.setBounds(0, 0, widthValue, heightValue);
		//show();
		setTitle(title);
		game.requestFocus();
		game.setPause(false);
		/* game.setSize(getSize().width - leftPanel.getSize().width,
			getSize().height); */
		//game.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : TertrisInterface.ICONS) {
			URL url = TertrisFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-bgColor={int}] [-sound={bool}] [-control={bool}]" + NL +
"\t[-level={starting_level}] [-units={size_of_polyiamond}]" + NL +
"\t[-columns={columns}] [-fill={rows}] [-fixedFill={bool}]" + NL +
"\t[-suddenAppear={bool}] [-diagonal={bool}] [-mixed={bool}]" + NL +
"\t[-bonus={bool}] [-cw={bool}] [-plain={bool}] [-showNext={bool}]" + NL +
"\t[-grid={bool}]");
			return;
		}
		TertrisFrame frame = new TertrisFrame(TertrisInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.game.requestFocusInWindow();
		frame.setLocation(TertrisFrame.xValue, TertrisFrame.yValue);
		frame.setVisible(true);
	}
}
